(** * Notations for functor categories *)
Require FunctorCategory.Core.
Require FunctorCategory.Morphisms.

Include FunctorCategory.Core.FunctorCategoryCoreNotations.
Include FunctorCategory.Morphisms.FunctorCategoryMorphismsNotations.
