(** * Kan Extensions *)
(** ** Definitions *)
Require KanExtensions.Core.
(** ** Kan Extensions assemble into functors *)
Require KanExtensions.Functors.

Include KanExtensions.Core.
Include KanExtensions.Functors.
