(** * Pointwise functors (functoriality of the functor category construction) *)
Require Functor.Pointwise.Core.
Require Functor.Pointwise.Properties.

Include Functor.Pointwise.Core.
Include Functor.Pointwise.Properties.
