(* Theory *)
Require Export WildCat.Adjoint.
Require Export WildCat.Core.
Require Export WildCat.Equiv.
Require Export WildCat.Opposite.
Require Export WildCat.Induced.
Require Export WildCat.EquivGpd.
Require Export WildCat.FunctorCat.
Require Export WildCat.NatTrans.
Require Export WildCat.Yoneda.
Require Export WildCat.Square.
Require Export WildCat.PointedCat.

(* Examples *)
Require Export WildCat.Universe.
Require Export WildCat.Paths.
Require Export WildCat.UnitCat.
Require Export WildCat.EmptyCat.
Require Export WildCat.Prod.
Require Export WildCat.Sum.
Require Export WildCat.Forall.
Require Export WildCat.Sigma.

(* Higher categories *)
Require Export WildCat.TwoOneCat.
