Require Adjoint.Composition.
Require Adjoint.Core.
Require Adjoint.Dual.
Require Adjoint.Identity.

Include Adjoint.Composition.AdjointCompositionNotations.
Include Adjoint.Core.AdjointCoreNotations.
Include Adjoint.Dual.AdjointDualNotations.
Include Adjoint.Identity.AdjointIdentityNotations.
